package biotransformerapi;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;

import org.apache.commons.lang3.StringUtils;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IAtomContainerSet;

import biotransformerapi.BiotransformerAPI.ioFormats;
import phase2filter.utils.FileUtils;


import phase2filter.prediction.P2Filter;

public class BiotransformerAPI {
	
	P2Filter p2filter = new P2Filter();
	LinkedHashMap<String, ArrayList<String>> validParams = new LinkedHashMap<String, ArrayList<String>>();
		

	
	public enum ioFormats{
		SDFILE, IATOMCONTAINERSET
	}
	
	
	public BiotransformerAPI() {

		ArrayList<String> required = new ArrayList<String>();
		ArrayList<String> optional = new ArrayList<String>();
		required.add("input");
		required.add("inputFormat");
		required.add("outputFormat");
		optional.add("output");
		this.validParams.put("required", required);
		this.validParams.put("optional", required);
	}
	

	
	
	
	public IAtomContainer predict(IAtomContainer molecule) throws Exception{
		boolean phaseII =this.p2filter.filter(molecule ).get(0) == "T";
		try {
			molecule.setProperty("Phase2Filter", StringUtils.capitalize(String.valueOf(phaseII)));	
		}
			
		catch (Exception e) {
			// Add a logger here
			System.err.println(e);
		}
		return molecule;	
	}
	
	public IAtomContainerSet predict(IAtomContainerSet molecules) throws Exception{
		IAtomContainerSet results = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainerSet.class);
		
		 for(IAtomContainer atc : molecules.atomContainers()) {
			 results.addAtomContainer(this.predict(atc));
		 }		
		return results;	
	}
	
	public Object predict(LinkedHashMap<String, Object>parameters) throws Exception {
		Object results 						= null;
		IAtomContainerSet inputMolecules 	= null;
		ioFormats inputFormat 				= ioFormats.valueOf( StringUtils.upperCase( (String) parameters.get("inputFormat") ))  ;
		ioFormats outputFormat 				= ioFormats.valueOf( StringUtils.upperCase( (String) parameters.get("outputFormat") ))  ;		
	
		
		if(this.validateParameters(parameters)) {
			if(inputFormat == ioFormats.IATOMCONTAINERSET) {
				inputMolecules = (IAtomContainerSet) parameters.get("input");
//				System.out.println("inputMolecules: " + inputMolecules.getAtomContainerCount());
			}
			else if(inputFormat == ioFormats.SDFILE) {
				inputMolecules = FileUtils.parseSdf(String.valueOf(parameters.get("input")) );
			}			
			
			IAtomContainerSet molsWithPredictions = null;
			try {
				molsWithPredictions = this.predict(inputMolecules);
//				System.out.println("molsWithPredictions: " + molsWithPredictions.getAtomContainerCount());
//				System.out.println( molsWithPredictions.getAtomContainer(0).getProperties());
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			if(outputFormat == ioFormats.IATOMCONTAINERSET) {
				results = molsWithPredictions;
			}
			else if(outputFormat == ioFormats.SDFILE) {
				String outputFileName = (String) parameters.get("output");
				
				if(molsWithPredictions != null) {
					FileUtils.saveAtomContainerSetToSDF(molsWithPredictions, outputFileName);
					results = outputFileName;
				}
				
			}			
		}
		
		return results;
	}

	public Boolean validateParameters(LinkedHashMap<String, Object> parameters) throws Exception {
		Boolean isValid = true;
		
		for(String requiredParam : this.validParams.get("required")) {
			if(!parameters.containsKey(requiredParam)) {
				throw new IOException("The required parameter \'" + requiredParam + "\' is missing. Please provide an adequate value.");
			}
		}
		
		
		ioFormats inputFormat 				= ioFormats.valueOf( StringUtils.upperCase( (String) parameters.get("inputFormat") ))  ;
		ioFormats outputFormat 				= ioFormats.valueOf( StringUtils.upperCase( (String) parameters.get("outputFormat") ))  ;	

		
		if(!Arrays.asList(ioFormats.values()).contains(inputFormat)) {
			isValid = false;
			throw new IOException("The provided input \'" + inputFormat + "\' format is not support by PhaseIIFilter");
		}
		if(!Arrays.asList(ioFormats.values()).contains(outputFormat)) {
			isValid = false;
			throw new IOException("The provided output \'" + inputFormat + "\' format is not support by PhaseIIFilter");
		}

		if(inputFormat == ioFormats.SDFILE) {
			File inputFile = new File((String) parameters.get("input"));
			if((!inputFile.exists()) || inputFile.isDirectory()) {
				isValid = false;
				throw new IOException("A pathname to the input file is missing. Please provide a pathname to the input file");			
				
			}		
		}
		
		if(outputFormat == ioFormats.SDFILE && !parameters.containsKey("output")) {
			isValid = false;
			throw new IOException("A pathname to the output file is missing. Please provide a pathname to the output file");			
		}
			
		return  isValid;							
	}
}
