package biotransformerapiphaseii;

import java.util.List;

import org.openscience.cdk.Bond;
import org.openscience.cdk.DefaultChemObjectBuilder;
import org.openscience.cdk.interfaces.IAtom;
import org.openscience.cdk.interfaces.IAtomContainer;
import org.openscience.cdk.interfaces.IBond;
import org.openscience.cdk.smiles.SmiFlavor;
import org.openscience.cdk.smiles.SmilesGenerator;
import org.openscience.cdk.smiles.SmilesParser;
import org.openscience.cdk.tools.manipulator.AtomContainerManipulator;

import phase2utils.UniqueIDFunctionSet;

public class GenerateMetaboliteStructure {
	SmilesParser sp = new SmilesParser(DefaultChemObjectBuilder.getInstance());
	SmilesGenerator sg = new SmilesGenerator(SmiFlavor.Isomeric);
	public IAtomContainer generateMetabolite(Integer atom_idx, IAtomContainer oneMole, String type) throws Exception{
		IAtomContainer result_struct = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainer.class);
		if(type.equals("UGT")) {
			result_struct = generateMetabolite_UGT(atom_idx, oneMole);				
		}
		else if(type.equals("SULF")) {
			result_struct = generateMetabolite_SULF(atom_idx, oneMole);
		}
		//System.out.println(this.sg.create(result_struct));
		return result_struct;
	}
	/**
	 * This function will take the UniqueID of the predicted reactive site and the original molecule, and generate(return) the corresponding UGT product
	 * If the predicted reactive site is invalid, it will return null.
	 * Reference for Glucuronidation: https://www.europeanmedical.info/metabolic-activation/glucuronide-formation.html
	 * @param atom_idx
	 * @param oneMole
	 * @return
	 * @throws Exception
	 */
	public IAtomContainer generateMetabolite_UGT(Integer atom_idx, IAtomContainer oneMole) throws Exception{
		IAtomContainer result_struct = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainer.class);
		String ugt_smiles = "[H]OC(=O)C1([H])O[C+]([H])C([H])(O[H])C([H])(O[H])C1([H])O[H]";
		IAtomContainer ugt_struct = this.sp.parseSmiles(ugt_smiles);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(ugt_struct);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(oneMole);
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
		result_struct.add(ugt_struct);
		result_struct.add(oneMole);
		for(int i = 0; i < ugt_struct.getAtomCount(); i++) {
			if(ugt_struct.getAtom(i).getFormalCharge() == 1) {
				IAtom connectAtom = ugt_struct.getAtom(i);
				IAtom acceptor_atom = UniqueIDFunctionSet.getAtomByUniqueID(atom_idx, oneMole);
				AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
				List<IAtom> neighborList = oneMole.getConnectedAtomsList(acceptor_atom);
				boolean processed = false;
				for(int j = 0; j < neighborList.size(); j++) {
					if(neighborList.get(j).getSymbol().equals("H")) {
						IAtom atomToRemove = neighborList.get(j);
						IBond bondToRemove = oneMole.getBond(acceptor_atom, atomToRemove);
						result_struct.removeAtom(atomToRemove);
						result_struct.removeBond(bondToRemove);
						processed = true;
						break;
					}
				}
				if(!processed && acceptor_atom.getSymbol().equals("N") && acceptor_atom.getFormalCharge() == 0) {
					acceptor_atom.setFormalCharge(acceptor_atom.getFormalCharge() + 1);
					processed = true;
				}
				//If cannot find any hydrogen, the return null. Glucuronidation requests to remove one H2O when the reaction occurs, so the acceptor must have a hydrogen attached except for Nitrogen.
				if(!processed) return null;
				connectAtom.setFormalCharge(connectAtom.getFormalCharge() - 1);
				IBond fakeBond = new Bond(acceptor_atom, connectAtom, IBond.Order.SINGLE);
				result_struct.addBond(fakeBond);
				AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(result_struct);
				AtomContainerManipulator.suppressHydrogens(result_struct);
				break;
			}
		}						
		
		//System.out.println(this.sg.create(result_struct));
		return result_struct;
	}
	
	/**
	 * This function will take the UniqueID of the predicted reactive site and the original molecule, and generate(return) the corresponding UGT product
	 * If the predicted reactive site is invalid, it will return null.
	 * Reference for Glucuronidation: https://www.europeanmedical.info/metabolic-activation/glucuronide-formation.html
	 * @param atom_idx
	 * @param oneMole
	 * @return
	 * @throws Exception
	 */
	public IAtomContainer generateMetabolite_SULF(Integer atom_idx, IAtomContainer oneMole) throws Exception{
		IAtomContainer result_struct = DefaultChemObjectBuilder.getInstance().newInstance(IAtomContainer.class);
		String ugt_smiles = "OS([O-])(=O)=O";
		IAtomContainer sulf_struct = this.sp.parseSmiles(ugt_smiles);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(sulf_struct);
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(oneMole);
		AtomContainerManipulator.convertImplicitToExplicitHydrogens(oneMole);
		result_struct.add(sulf_struct);
		result_struct.add(oneMole);
		for(int i = 0; i < sulf_struct.getAtomCount(); i++) {
			//Find the oxgen to remove
			if(sulf_struct.getAtom(i).getFormalCharge() == -1) {
				IAtom atomToRemove = sulf_struct.getAtom(i);
				List<IAtom> atomToConnectList = sulf_struct.getConnectedAtomsList(atomToRemove);
				//In this process, the O- atom should only connect to the Sulfur atom
				if(atomToConnectList.size() > 1) return null;
				IAtom atomToConnect = atomToConnectList.get(0);
				IAtom acceptor_atom = UniqueIDFunctionSet.getAtomByUniqueID(atom_idx, oneMole);
				IBond bondToRemove = result_struct.getBond(atomToRemove, atomToConnect);
				result_struct.removeAtom(atomToRemove);
				result_struct.removeBond(bondToRemove);
				List<IAtom> acceptor_atom_neighborList = result_struct.getConnectedAtomsList(acceptor_atom);
				//Remove the connected hydrogen atom and add a bond to connect the sulfate group and the substrate
				for(int j = 0; j < acceptor_atom_neighborList.size(); j++) {
					if(acceptor_atom_neighborList.get(j).getSymbol().equals("H")) {
						IAtom hydrogen_to_remove = acceptor_atom_neighborList.get(j);
						IBond bondToRemove_hydrogen = result_struct.getBond(acceptor_atom, hydrogen_to_remove);
						result_struct.removeAtom(hydrogen_to_remove);
						result_struct.removeBond(bondToRemove_hydrogen);
						break;
					}
				}
				IBond newBond = new Bond(acceptor_atom, atomToConnect, IBond.Order.SINGLE);
				result_struct.addBond(newBond);
				break;
			}
		}						
		AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(result_struct);
		AtomContainerManipulator.suppressHydrogens(result_struct);
		return result_struct;
	}
}
